"use strict";

exports.__esModule = true;
exports.safeSpecTransforms = exports.applyNonTransformProps = exports.removeMagicProps = exports.onlySpecTransforms = exports.makePruningSpec = exports.makeFunctionUnlessObject = exports.alwaysFunction = exports.objectify = exports.isPlainObj = void 0;

var _ramda = require("ramda");

/**
 * This is what you really mean when you're checking to see if something is an Object.
 * Honestly who cares that an Array, Date and RegExp are technically objects too?
 * You were looking for the thing with the curly braces {} right?
 *
 * @func
 * @sig * -> Boolean
 * @param {*} val A value of any type
 * @returns {Boolean} Whether or not the value is an Object
 */
var isPlainObj = (0, _ramda.pathEq)(['constructor', 'name'], 'Object');
/**
 * Turns any non-Object values into empty objects
 *
 * @func
 * @sig * -> {k: v}
 * @param {*} val A value of any type
 * @returns {Object} If the original value was an Object, it is returned as-is,
 * otherwise an empty object is returned
 */

exports.isPlainObj = isPlainObj;
var objectify = (0, _ramda.unless)(isPlainObj, (0, _ramda.always)({}));
/**
 * If the provided value is NOT a function, transform it into a function that
 * always returns that value
 *
 * @func
 * @sig * -> (* -> *)
 * @param {*} val A value of any type
 * @returns {Function} If the original value was a function, that function is returned,
 * otherwise a function that always returns the original value is returned
 * function that always returns that value OR
 */

exports.objectify = objectify;
var alwaysFunction = (0, _ramda.unless)((0, _ramda.is)(Function), _ramda.always);
/**
 * If the provided value is neither an Object nor a Function, transform it into
 * a function that always returns that provided value.
 *
 * @func
 * @sig * -> (* -> *)|{k: v}
 * @param {*} val A value of any type
 * @returns {Function|Object} A function that always returns the provided value
 * OR the original value (if it was already a Function or an Object)
 */

exports.alwaysFunction = alwaysFunction;
var makeFunctionUnlessObject = (0, _ramda.unless)((0, _ramda.either)((0, _ramda.is)(Function), isPlainObj), _ramda.always);
/**
 * Makes sure a given spec is acceptable for one of the pruning modes (ie, "remove" or "keep").
 * Specs in this mode are treated more strictly, meaning the prop must be given
 * a value of "true" or the key and value can be identical.
 * Functions are acceptable values in the spec of course.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec A spec to be coerced into an acceptable pruning spec
 * @returns {Object} A spec that is acceptable to be used in pruning mode.
 */

exports.makeFunctionUnlessObject = makeFunctionUnlessObject;
var makePruningSpec = (0, _ramda.pipe)(objectify, _ramda.toPairs, (0, _ramda.filter)(function (_ref) {
  var key = _ref[0],
      val = _ref[1];
  return key === val || val === true || (0, _ramda.is)(Function, val);
}), _ramda.fromPairs);
/**
 * Filters a spec (object) to only the props that are transform functions. If
 * the input is not an object, this is just an identity function.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec An object whose values (may) be transform functions
 * @returns {Object} The input object with only the props that are functions retained
 */

exports.makePruningSpec = makePruningSpec;
var onlySpecTransforms = (0, _ramda.when)(isPlainObj, (0, _ramda.pipe)(_ramda.toPairs, (0, _ramda.filter)(function (_ref2) {
  var key = _ref2[0],
      val = _ref2[1];
  return key === 'shapeyDebug' || (0, _ramda.is)(Function, val);
}), _ramda.fromPairs));
/**
 * Removes the reserved "shapey*" prefixed props from a spec.
 *
 * @func
 * @sig {k: v} -> {k: v}
 * @param {Object} spec A shapey spec to be pruned of magic props
 * @returns {Object} A shapey spec cleaned of magic props
 */

exports.onlySpecTransforms = onlySpecTransforms;
var removeMagicProps = (0, _ramda.pipe)(_ramda.toPairs, (0, _ramda.reject)(function (_ref3) {
  var key = _ref3[0];
  return (0, _ramda.test)(/^shapey/i, key);
}), _ramda.fromPairs);
/**
 * Takes an non-transform props (non-functions) in a given spec and merges them onto the input object.
 *
 * @func
 * @sig {k: v} -> {k: v} -> {k: v}
 * @param {Object} spec An object whose values are either transform functions or pass-through values to be added to the return object
 * @param {Object} input An object which will be passed through the re-shaping transform functions defined by the spec
 * @returns {Object} The input object with only the pass-through props applied to it
 */

exports.removeMagicProps = removeMagicProps;
var applyNonTransformProps = (0, _ramda.curry)(function (spec, value) {
  return [spec, value].every(isPlainObj) ? (0, _ramda.pipe)(removeMagicProps, (0, _ramda.reject)((0, _ramda.is)(Function)), (0, _ramda.merge)(value))(spec) : value;
});
/**
 * Logs to the console any failed transform functions, along with the field name
 * and the value that was fed into the transform function. Plus, the exception
 * itself is also logged.
 *
 * @func
 * @sig String -> * -> Error -> *
 * @param {String} fieldName The field name for the failed transform function
 * @param {Object} value The value that was fed into the transform function
 * @param {Object} exception The exception thrown by the failed transform
 * @returns {undefined}
 */

exports.applyNonTransformProps = applyNonTransformProps;
var defaultErrorHandler = (0, _ramda.curry)(function (fieldName, value, exception) {
  return (// eslint-disable-next-line no-console
    console.error("\n Transform failed on field: \"" + fieldName + "\"\n value:", value, '\n', exception)
  );
});
/**
 * A wrapper around an (optional) error handler that the may be passed in as the
 * "shapeyDebug" value. It is curried, but will feed their handler a more
 * standard signature of: (err, field, value)
 *
 * @func
 * @sig ((Error, String, *) -> *) -> String -> * -> Error -> *
 * @param {Function} errHandler The custom error handler provided by the consumer
 * @param {String} fieldName The field name for the failed transform function
 * @param {Object} value The value that was fed into the transform function
 * @param {Object} exception The exception thrown by the failed transform
 * @returns {*} Could be anything, or nothing (it's up to the consumer)
 */

var wrapperForTheirErrorHandler = (0, _ramda.curry)(function (errHandler, fieldName, value, exception) {
  return errHandler(exception, fieldName, value);
});
/**
 * Wraps every function on a given spec in a try/catch that catches any exception.
 *
 * What it _does_ with the exception is up to the consumer:
 *   - ignores it (if "shapeyDebug" isn't set)
 *   - logs it along with the corresponding field name,
 *     using `console.error` (if "shapeyDebug" is set to `true`)
 *   - uses a custom handler supplied by the consumer as the value for "shapeyDebug"
 *
 * Note that a custom error handler will be passed the following params (in order):
 *   - The exception
 *   - The field name
 *   - The value that was fed into the transform function
 *
 * @func
 * @sig {k: (a -> b)} -> {k: (a -> b)}
 * @param {Object} spec An object whose values may be transform functions that
 * need to be wrapped in try/catch
 * @returns {Object} The same spec object but whose functions are
 * safely wrapped in in try/catch handlers
 */

var safeSpecTransforms = (0, _ramda.pipe)(objectify, (0, _ramda.converge)((0, _ramda.assoc)('shapeyDebug'), [(0, _ramda.pipe)((0, _ramda.prop)('shapeyDebug'), (0, _ramda.cond)([[(0, _ramda.equals)(true), (0, _ramda.always)(defaultErrorHandler)], [(0, _ramda.is)(Function), wrapperForTheirErrorHandler], [(0, _ramda.test)(/^skip$/i), function () {
  return function (_, originalValue) {
    return (0, _ramda.always)(originalValue);
  };
}], [_ramda.T, function () {
  return function () {
    return (0, _ramda.always)(undefined);
  };
}]])), _ramda.identity]), (0, _ramda.mapObjIndexed)(function (transform, fieldName, obj) {
  return (0, _ramda.when)((0, _ramda.is)(Function), (0, _ramda.curry)(function (fn, val) {
    return (0, _ramda.tryCatch)(fn, obj.shapeyDebug(fieldName, val))(val);
  }))(transform);
}), (0, _ramda.omit)(['shapeyDebug']));
exports.safeSpecTransforms = safeSpecTransforms;